#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Compatibility shim for Kilo, while operators migrate to oslo.middleware."""

from oslo_middleware import catch_errors

from networking_odl.openstack.common import versionutils


@versionutils.deprecated(as_of=versionutils.deprecated.KILO,
                         in_favor_of='oslo_middleware.CatchErrors')
class CatchErrorsMiddleware(catch_errors.CatchErrors):
    pass
